# SPDX-License-Identifier: Apache-2.0

"""
The io module holds the abstraction and implementations for Granite 3.2 IO
processing.
"""

# Local
import granite_io.io.granite_3_2.granite_3_2  # noqa: F401
