# SPDX-License-Identifier: Apache-2.0

"""
The input processor module holds the abstraction and implementations for
Granite 3.3 input processing.
"""

# Local
import granite_io.io.granite_3_3.input_processors.granite_3_3_input_processor  # noqa: F401
